/*-------------------------------------------------------------------------
 *
 * boxproc.c--
 *    pg_amproc entries for GiSTs over 2-D boxes.
 * This gives R-tree behavior, with Guttman's poly-time split algorithm.
 *
 *
 *
 * IDENTIFICATION
 *
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include "access/gist.h"
#include "access/itup.h"
#include "access/rtree.h"

#include "utils/palloc.h"
#include "utils/geo_decls.h"
#include "utils/elog.h"

bool gbox_leaf_consistent(BOX *key, BOX *query, StrategyNumber strategy);
bool gbox_internal_consistent(BOX *key,	BOX *query, StrategyNumber strategy);
BOX *gbox_binary_union(BOX *r1, BOX *r2, int *sizep);
BOX *gbox_union(bytea *entryvec, int *sizep);
bool gbox_consistent(GISTENTRY *entry,BOX *query,StrategyNumber strategy);
GISTENTRY * gbox_compress(GISTENTRY *entry);
GISTENTRY * gbox_decompress(GISTENTRY *entry);
float * gbox_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result);
GIST_SPLITVEC * gbox_picksplit(bytea *entryvec, GIST_SPLITVEC *v);
bool * gbox_same(BOX *b1, BOX *b2, bool *result);
bool gbox_leaf_consistent(BOX *key, BOX *query, StrategyNumber strategy);
bool gbox_internal_consistent(BOX *key, BOX *query, StrategyNumber strategy);


/* used for sorting boxes */
typedef struct boxix {
    int index;
    BOX b;
} BOXIX;

/*
** The GiST Consistent method for boxes
** Should return false if for all data items x below entry,
** the predicate x op query == FALSE, where op is the oper
** corresponding to strategy in the pg_amop table.
*/
bool 
gbox_consistent(GISTENTRY *entry,
	       BOX *query,
	       StrategyNumber strategy)
{
    /*
    ** if entry is not leaf, use gbox_internal_consistent,
    ** else use gbox_leaf_consistent
    */
    if (GIST_LEAF(entry))
      return(gbox_leaf_consistent((BOX *)(entry->pred), query, strategy));
    else
      return(gbox_internal_consistent((BOX *)(entry->pred), query, strategy));
}


/*
** The GiST Union method for boxes
** returns the minimal bounding box that encloses all the entries in entryvec
*/
BOX *
gbox_union(bytea *entryvec, int *sizep)
{
    int numranges, i;
    BOX *out, *tmp;

    numranges = (VARSIZE(entryvec) - VARHDRSZ)/sizeof(GISTENTRY); 
    tmp = (BOX *)(((GISTENTRY *)(VARDATA(entryvec)))[0]).pred;
    *sizep = sizeof(BOX);
    out = NULL;

    for (i = 1; i < numranges; i++) {
	out = gbox_binary_union(tmp, (BOX *)
				 (((GISTENTRY *)(VARDATA(entryvec)))[i]).pred,
				 sizep);
	if (i > 1) pfree(tmp);
	tmp = out;
    }

    return(out);
}

/*
** GiST Compress and Decompress methods for boxes
** do not do anything.
*/
GISTENTRY *
gbox_compress(GISTENTRY *entry)
{
    return(entry);
}

GISTENTRY *
gbox_decompress(GISTENTRY *entry)
{
    return(entry);
}

/*
** The GiST Penalty method for boxes
** As in the R-tree paper, we use change in area as our penalty metric
*/
float *
gbox_penalty(GISTENTRY *origentry, GISTENTRY *newentry, float *result)
{
    char * ud;
    float tmp1, tmp2;
    
    ud = DatumGetPointer( DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( origentry->pred ),
		PointerGetDatum( newentry->pred )) );
    DirectFunctionCall2( rt_box_size,
	PointerGetDatum( ud ),
	PointerGetDatum( &tmp1 ) );
    DirectFunctionCall2( rt_box_size,
	PointerGetDatum( origentry->pred ),
	PointerGetDatum( &tmp2 ) );
    *result = tmp1 - tmp2;
    pfree((char *)ud);
    return(result);
}



/*
** The GiST PickSplit method for boxes
** We use Guttman's poly time split algorithm 
*/
GIST_SPLITVEC *
gbox_picksplit(bytea *entryvec,
	      GIST_SPLITVEC *v)
{
    OffsetNumber i, j;
    char *datum_alpha, *datum_beta;
    char *datum_l, *datum_r;
    char *union_d, *union_dl, *union_dr;
    char *inter_d;
    bool firsttime;
    float size_alpha, size_beta, size_union, size_inter;
    float size_waste, waste;
    float size_l, size_r;
    int nbytes;
    OffsetNumber seed_1 = 0, seed_2 = 0;
    OffsetNumber *left, *right;
    OffsetNumber maxoff;

    maxoff = ((VARSIZE(entryvec) - VARHDRSZ)/sizeof(GISTENTRY)) - 2;
    nbytes =  (maxoff + 2) * sizeof(OffsetNumber);
    v->spl_left = (OffsetNumber *) palloc(nbytes);
    v->spl_right = (OffsetNumber *) palloc(nbytes);
    
    firsttime = true;
    waste = 0.0;
    
    for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i)) {
	datum_alpha = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[i].pred);
	for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j)) {
	    datum_beta = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[j].pred);
	    
	    /* compute the wasted space by unioning these guys */
	    /* size_waste = size_union - size_inter; */
    	    union_d = DatumGetPointer(DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( datum_alpha ),
		PointerGetDatum( datum_beta )) );
    	    DirectFunctionCall2( rt_box_size,
		PointerGetDatum( union_d ),
		PointerGetDatum( &size_union ) );
    	    inter_d = DatumGetPointer(DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( datum_alpha ),
		PointerGetDatum( datum_beta )) );
    	    DirectFunctionCall2( rt_box_size,
		PointerGetDatum( inter_d ),
		PointerGetDatum( &size_inter ) );
	    size_waste = size_union - size_inter;
	    
	    pfree(union_d);
	    
	    if (inter_d != (char *) NULL)
		pfree(inter_d);
	    
	    /*
	     *  are these a more promising split that what we've
	     *  already seen?
	     */
	    
	    if (size_waste > waste || firsttime) {
		waste = size_waste;
		seed_1 = i;
		seed_2 = j;
		firsttime = false;
	    }
	}
    }
    
    left = v->spl_left;
    v->spl_nleft = 0;
    right = v->spl_right;
    v->spl_nright = 0;
    
    datum_alpha = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[seed_1].pred);
    datum_l = DatumGetPointer( DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( datum_alpha ),
		PointerGetDatum( datum_alpha )) );
    DirectFunctionCall2( rt_box_size,
	PointerGetDatum( datum_l ),
	PointerGetDatum( &size_l ) );
    datum_beta = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[seed_2].pred);;
    datum_r = DatumGetPointer( DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( datum_beta ),
		PointerGetDatum( datum_beta )) );
    DirectFunctionCall2( rt_box_size,
	PointerGetDatum( datum_r ),
	PointerGetDatum( &size_r ) );
    
    /*
     *  Now split up the regions between the two seeds.  An important
     *  property of this split algorithm is that the split vector v
     *  has the indices of items to be split in order in its left and
     *  right vectors.  We exploit this property by doing a merge in
     *  the code that actually splits the page.
     *
     *  For efficiency, we also place the new index tuple in this loop.
     *  This is handled at the very end, when we have placed all the
     *  existing tuples and i == maxoff + 1.
     */
    
    maxoff = OffsetNumberNext(maxoff);
    for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i)) {
	
	/*
	 *  If we've already decided where to place this item, just
	 *  put it on the right list.  Otherwise, we need to figure
	 *  out which page needs the least enlargement in order to
	 *  store the item.
	 */
	
	if (i == seed_1) {
	    *left++ = i;
	    v->spl_nleft++;
	    continue;
	} else if (i == seed_2) {
	    *right++ = i;
	    v->spl_nright++;
	    continue;
	}
	
	/* okay, which page needs least enlargement? */ 
	datum_alpha = (char *)(((GISTENTRY *)(VARDATA(entryvec)))[i].pred);
        union_dl = DatumGetPointer( DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( datum_l ),
		PointerGetDatum( datum_alpha )) );
        union_dr = DatumGetPointer( DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( datum_r ),
		PointerGetDatum( datum_alpha )) );
    	DirectFunctionCall2( rt_box_size,
		PointerGetDatum( union_dl ),
		PointerGetDatum( &size_alpha ) );
    	DirectFunctionCall2( rt_box_size,
		PointerGetDatum( union_dr ),
		PointerGetDatum( &size_beta ) );
	
	/* pick which page to add it to */
	if (size_alpha - size_l < size_beta - size_r) {
	    pfree(datum_l);
	    pfree(union_dr);
	    datum_l = union_dl;
	    size_l = size_alpha;
	    *left++ = i;
	    v->spl_nleft++;
	} else {
	    pfree(datum_r);
	    pfree(union_dl);
	    datum_r = union_dr;
	    size_r = size_alpha;
	    *right++ = i;
	    v->spl_nright++;
	}
    }
    *left = *right = FirstOffsetNumber;	/* sentinel value, see dosplit() */
    
    v->spl_ldatum = datum_l;
    v->spl_rdatum = datum_r;

    return v;
}

/*
** Equality method
*/
bool *
gbox_same(BOX *b1, BOX *b2, bool *result)
{
  if (DatumGetBool( DirectFunctionCall2( box_same, PointerGetDatum(b1), PointerGetDatum(b2)) ))
    *result = TRUE;
  else *result = FALSE;
  return(result);
}

/* 
** SUPPORT ROUTINES
*/
bool 
gbox_leaf_consistent(BOX *key,
		     BOX *query,
		     StrategyNumber strategy)
{
    bool retval;

    switch(strategy) {
    case RTLeftStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_left, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTOverLeftStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overleft, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTOverlapStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overlap, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTOverRightStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overright, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTRightStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_right, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTSameStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_same, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTContainsStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_contain, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTContainedByStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_contained, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    default:
      retval = FALSE;
    }
    return(retval);
}

bool 
gbox_internal_consistent(BOX *key,
			BOX *query,
			StrategyNumber strategy)
{
    bool retval;

    switch(strategy) {
    case RTLeftStrategyNumber:
    case RTOverLeftStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overleft, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTOverlapStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overlap, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTOverRightStrategyNumber:
    case RTRightStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_right, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTSameStrategyNumber:
    case RTContainsStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_contain, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    case RTContainedByStrategyNumber:
      retval = DatumGetBool( DirectFunctionCall2( box_overlap, PointerGetDatum(key), PointerGetDatum(query) ) );
      break;
    default:
      retval = FALSE;
    }
    return(retval);
}

BOX *
gbox_binary_union(BOX *r1, BOX *r2, int *sizep)
{
    BOX *retval;

    retval = (BOX*)DatumGetPointer( DirectFunctionCall2(
		rt_box_union,
		PointerGetDatum( r1 ),
		PointerGetDatum( r2 )) );
    *sizep = sizeof(BOX);

    return retval;
}

